@@ -2,7 +2,7 @@
 
 /*  mcall.c -- multiallelic and rare variant calling.
 
-    Copyright (C) 2012-2021 Genome Research Ltd.
+    Copyright (C) 2012-2022 Genome Research Ltd.
 
     Author: Petr Danecek <pd3@sanger.ac.uk>
 
@@ -316,7 +316,7 @@ static void init_sample_groups(call_t *call)
             while ( *ptr && isspace(*ptr) ) ptr++;
             if ( !*ptr ) error("Could not parse the line in %s, expected a sample name followed by tab and a population name: %s\n",call->sample_groups,lines[i]);
             *tmp = 0;
-            int ismpl = bcf_hdr_id2int(call->hdr, BCF_DT_SAMPLE, lines[i]); 
+            int ismpl = bcf_hdr_id2int(call->hdr, BCF_DT_SAMPLE, lines[i]);
             if ( ismpl<0 ) continue;
             if ( smpl2grp[ismpl] ) error("Error: the sample \"%s\" is listed twice in %s\n", lines[i],call->sample_groups);
             if ( !khash_str2int_has_key(grp2idx,ptr+1) )
@@ -338,7 +338,7 @@ static void init_sample_groups(call_t *call)
         {
             if ( !smpl2grp[i] ) error("Error: The sample \"%s\" is not listed in %s\n",call->hdr->samples[i],call->sample_groups);
             int igrp = smpl2grp[i] - 1;
-            if ( !call->smpl_grp[igrp].nsmpl ) 
+            if ( !call->smpl_grp[igrp].nsmpl )
                 call->smpl_grp[igrp].smpl = (uint32_t*)calloc(grp2n[igrp],sizeof(uint32_t));
             call->smpl_grp[igrp].smpl[call->smpl_grp[igrp].nsmpl] = i;
             call->smpl_grp[igrp].nsmpl++;
@@ -446,7 +446,7 @@ void mcall_destroy(call_t *call)
 // qual calculation is not affected.
 // Missing values are replaced by generic likelihoods when X (unseen allele) is
 // present.
-// NB: While the -m callig model uses the pdgs in canonical order,
+// NB: While the -m calling model uses the pdgs in canonical order,
 // the original samtools -c calling code uses pdgs in reverse order (AA comes
 // first, RR last).
 // NB: Ploidy is not taken into account here, which is incorrect.
@@ -747,7 +747,7 @@ static void mcall_set_ref_genotypes(call_t *call, int nals_ori)
 static void mcall_call_genotypes(call_t *call, int nals_ori, smpl_grp_t *grp)
 {
     int ia, ib, i;
-    int ngts_ori = nals_ori*(nals_ori+1)/2; 
+    int ngts_ori = nals_ori*(nals_ori+1)/2;
     int ngts_new = call->nals_new*(call->nals_new+1)/2;
     int nsmpl = grp->nsmpl;
 
@@ -1273,8 +1273,9 @@ void mcall_trim_and_update_numberR(call_t *call, bcf1_t *rec, int nals_ori, int
 static int mcall_constrain_alleles(call_t *call, bcf1_t *rec, int *unseen)
 {
     assert( call->tgt_als->n );
-    if ( call->tgt_als->n>5 ) error("Maximum accepted number of alleles is 5, got %d\n", call->tgt_als->n);
     hts_expand(char*,call->tgt_als->n+1,call->nals,call->als);
+    hts_expand(int,call->tgt_als->n+1,call->nals_map,call->als_map);
+    hts_expand(int,(call->tgt_als->n+1)*(call->tgt_als->n+2)/2,call->npl_map,call->pl_map);
 
     int has_new = 0;
 
@@ -1292,18 +1293,6 @@ static int mcall_constrain_alleles(call_t *call, bcf1_t *rec, int *unseen)
     {
         call->als[nals] = call->tgt_als->allele[i];
         j = vcmp_find_allele(call->vcmp, rec->d.allele+1, rec->n_allele - 1, call->tgt_als->allele[i]);
-
-        // if ( j+1==*unseen )
-        // { 
-        //     fprintf(bcftools_stderr,"Fixme? Cannot constrain to %d-th allele (%s); j=%d,unseen=%d. VCF=",i,call->tgt_als->allele[i],j,*unseen);
-        //     int k;
-        //     for (k=0; k<rec->n_allele; k++) fprintf(bcftools_stderr,"%s%s",k==0?"":",",rec->d.allele[k]);
-        //     fprintf(bcftools_stderr,"\tTAB=");
-        //     for (k=0; k<call->tgt_als->n; k++) fprintf(bcftools_stderr,"%s%s",k==0?"":",",call->tgt_als->allele[k]);
-        //     fprintf(bcftools_stderr,"\n");
-        //     return -1;
-        // }
-        
         if ( j>=0 )
         {
             // existing allele
@@ -1508,7 +1497,7 @@ int mcall(call_t *call, bcf1_t *rec)
     // If available, take into account reference panel AFs
     if ( call->prior_AN && bcf_get_info_int32(call->hdr, rec, call->prior_AN ,&call->ac, &call->nac)==1 )
     {
-        int an = call->ac[0];   // number of alleles total, procede only if not zero; reuse call->ac
+        int an = call->ac[0];   // number of alleles total, proceed only if not zero; reuse call->ac
         if ( an > 0 && bcf_get_info_int32(call->hdr, rec, call->prior_AC ,&call->ac, &call->nac)==nals_ori-1 )    // number of ALT alleles
         {
             int ac0 = an;       // this will become the number of REFs
@@ -1539,9 +1528,9 @@ int mcall(call_t *call, bcf1_t *rec)
     bcf_update_info_int32(call->hdr, rec, "QS", NULL, 0);      // remove QS tag
 
     if ( nals_ori > 8*sizeof(call->als_new) )
-    { 
+    {
         fprintf(bcftools_stderr,"Too many alleles at %s:%"PRId64", skipping.\n", bcf_seqname(call->hdr,rec),(int64_t) rec->pos+1);
-        return 0; 
+        return 0;
     }
 
     // For each group find the best combination of alleles
@@ -1571,6 +1560,11 @@ int mcall(call_t *call, bcf1_t *rec)
     call->nals_new = 0;
     for (i=0; i<nals_ori; i++)
     {
+        if ( (call->flag&CALL_KEEP_UNSEEN) && i==unseen && call->nals_new==1 )
+        {
+            call->nals_new++;
+            call->als_new |= 1<<i;
+        }
         if ( i>0 && i==unseen ) continue;
         if ( call->flag & CALL_KEEPALT ) call->als_new |= 1<<i;
         if ( call->als_new & (1<<i) ) call->nals_new++;
@@ -1598,9 +1592,9 @@ int mcall(call_t *call, bcf1_t *rec)
         for (i=0; i<call->nals_new; i++) call->ac[i] = 0;
 
         if ( call->flag & CALL_CONSTR_TRIO && call->nals_new>4 )
-        { 
+        {
             fprintf(bcftools_stderr,"Too many alleles at %s:%"PRId64", skipping.\n", bcf_seqname(call->hdr,rec),(int64_t) rec->pos+1);
-            return 0; 
+            return 0;
         }
         if ( call->output_tags & (CALL_FMT_GQ|CALL_FMT_GP) )
         {
@@ -1672,7 +1666,7 @@ int mcall(call_t *call, bcf1_t *rec)
             anno16_t a;
             float tmpf[4];
             int is_tested = test16(call->anno16, &a) >= 0 && a.is_tested ? 1 : 0;
-            if ( is_tested ) 
+            if ( is_tested )
             {
                 for (i=0; i<4; i++) tmpf[i] = a.p[i];
                 bcf_update_info_float(call->hdr, rec, "PV4", tmpf, 4);
@@ -1682,6 +1676,6 @@ int mcall(call_t *call, bcf1_t *rec)
 
     bcf_update_info_int32(call->hdr, rec, "I16", NULL, 0);     // remove I16 tag
 
-    return call->nals_new;
+    return is_variant ? call->nals_new : 1;
 }
 
