@@ -1,7 +1,7 @@
 /*  bedidx.c -- BED file indexing.
 
     Copyright (C) 2011 Broad Institute.
-    Copyright (C) 2014,2017 Genome Research Ltd.
+    Copyright (C) 2014, 2017-2019, 2024 Genome Research Ltd.
 
     Author: Heng Li <lh3@sanger.ac.uk>
 
@@ -34,28 +34,31 @@ DEALINGS IN THE SOFTWARE.  */
 #include "bedidx.h"
 
 #include "htslib/ksort.h"
-KSORT_INIT_GENERIC(uint64_t)
 
 #include "htslib/kseq.h"
 KSTREAM_INIT(gzFile, gzread, 8192)
 
+static inline int lt_pair_pos(hts_pair_pos_t a, hts_pair_pos_t b) {
+    if (a.beg == b.beg) return a.end < b.end;
+    return a.beg < b.beg;
+}
+KSORT_INIT_STATIC(hts_pair_pos_t, hts_pair_pos_t, lt_pair_pos)
+
 /*! @typedef
  * @abstract bed_reglist_t - value type of the BED hash table
  * This structure encodes the list of intervals (ranges) for the regions provided via BED file or
  * command line arguments.
- * @field *a           pointer to the array of intervals (kept as 64 bit integers). The upper 32 bits
- * encode the beginning of the interval, while the lower 32 bits encode the end, for easy sorting.
- * |-- 32 bits --|-- 32 bits --|
- * |---- beg ----|---- end ----|
+ * @field *a           pointer to the array of intervals.
  * @field n            actual number of elements contained by a
  * @field m            number of allocated elements to a (n <= m)
  * @field *idx         index array for computing the minimum offset
  */
 typedef struct {
     int n, m;
-    uint64_t *a;
+    hts_pair_pos_t *a;
     int *idx;
     int filter;
+    hts_pos_t max_idx;
 } bed_reglist_t;
 
 #include "htslib/khash.h"
@@ -71,7 +74,6 @@ static void bed_print(void *reg_hash) {
     khint_t k;
     int i;
     const char *reg;
-    uint32_t beg, end;
 
     if (!h) {
         printf("Hash table is empty!\n");
@@ -84,10 +86,8 @@ static void bed_print(void *reg_hash) {
             if ((p = &kh_val(h,k)) != NULL && p->n > 0) {
                 printf("Filter: %d\n", p->filter);
                 for (i=0; i<p->n; i++) {
-                    beg = (uint32_t)(p->a[i]>>32);
-                    end = (uint32_t)(p->a[i]);
-
-                    printf("\tinterval[%d]: %d-%d\n",i,beg,end);
+                    printf("\tinterval[%d]: %"PRIhts_pos"-%"PRIhts_pos"\n",
+                           i,p->a[i].beg,p->a[i].end);
                 }
             } else {
                 printf("Region '%s' has no intervals!\n", reg);
@@ -97,80 +97,88 @@ static void bed_print(void *reg_hash) {
 }
 #endif
 
-static int *bed_index_core(int n, uint64_t *a)
+static int bed_index_core(bed_reglist_t *regions)
 {
-    int i, j, l, *idx;
-    l = 0; idx = 0;
-    for (i = 0; i < n; ++i) {
-        int beg, end;
-        beg = a[i]>>32 >> LIDX_SHIFT; end = ((uint32_t)a[i]) >> LIDX_SHIFT;
-        if (l < end + 1) {
-            int old_l = l;
-            l = end + 1;
-            kroundup32(l);
-            idx = realloc(idx, l * sizeof(int));
-            if (!idx)
-                return NULL;
+    int i, *idx = NULL;
+    size_t idx_size = 0;
+    hts_pos_t last_end = 0;
+    hts_pair_pos_t *a = regions->a;
 
-            for (j = old_l; j < l; ++j)
-                idx[j] = -1;
+    // Construct a linear index on regions, to allow rapid lookup of
+    // where to start searching for matches
+    for (i = 0; i < regions->n; ++i) {
+        hts_pos_t beg = a[i].beg >= 0 ? a[i].beg >> LIDX_SHIFT : 0;
+        hts_pos_t end = a[i].end >= 0 ? a[i].end >> LIDX_SHIFT : 0;
+        hts_pos_t j;
+        if (end < last_end)
+            continue;  // Can happen for a containment
+        if (end + 1 >= SIZE_MAX / sizeof(*idx)) { // Ensure no overflow
+            errno = ENOMEM;
+            free(idx);
+            return -1;
         }
-
-        for (j = beg; j < end+1; ++j)
-            if (idx[j] < 0)
-                idx[j] = i;
+        if (hts_resize(int, (size_t) end + 1, &idx_size, &idx, 0) < 0) {
+            free(idx);
+            return -1;
+        }
+        // Fill any gap prior to this region by pointing to the previous one
+        for (j = last_end; j < beg; j++)
+            idx[j] = i > 0 ? i - 1 : 0;
+        // Fill from max(last_end, beg) to `end` (inclusive) with current region
+        for (; j <= end; j++)
+            idx[j] = i;
+        // Remember where finished for the next gap
+        last_end = end + 1;
     }
-    return idx;
+    regions->idx = idx;
+    regions->max_idx = last_end;
+    return 0;
 }
 
-static void bed_index(void *_h)
+static int bed_index(reghash_t *h)
 {
-    reghash_t *h = (reghash_t*)_h;
     khint_t k;
     for (k = 0; k < kh_end(h); ++k) {
         if (kh_exist(h, k)) {
             bed_reglist_t *p = &kh_val(h, k);
-            if (p->idx) free(p->idx);
-            ks_introsort(uint64_t, p->n, p->a);
-            p->idx = bed_index_core(p->n, p->a);
+            if (p->idx) {
+                free(p->idx);
+                p->idx = NULL;
+            }
+            ks_introsort(hts_pair_pos_t, p->n, p->a);
+            if (bed_index_core(p) != 0) {
+                return -1;
+            }
         }
     }
+    return 0;
 }
 
-static int bed_minoff(const bed_reglist_t *p, unsigned int beg, unsigned int end) {
-    int i, min_off=0;
+static int bed_minoff(const bed_reglist_t *p, hts_pos_t beg) {
+    int min_off=0;
 
-    if (p && p->idx) {
-        min_off = (beg>>LIDX_SHIFT >= p->n)? p->idx[p->n-1] : p->idx[beg>>LIDX_SHIFT];
-        if (min_off < 0) { // TODO: this block can be improved, but speed should not matter too much here
-            int n = beg>>LIDX_SHIFT;
-            if (n > p->n)
-                n = p->n;
-            for (i = n - 1; i >= 0; --i)
-                if (p->idx[i] >= 0)
-                    break;
-            min_off = i >= 0? p->idx[i] : 0;
-        }
+    if (p && p->idx && p->max_idx > 0 && beg >= 0) {
+        min_off = (beg>>LIDX_SHIFT >= p->max_idx)? p->idx[p->max_idx-1] : p->idx[beg>>LIDX_SHIFT];
     }
 
     return min_off;
 }
 
-static int bed_overlap_core(const bed_reglist_t *p, int beg, int end)
+static int bed_overlap_core(const bed_reglist_t *p, hts_pos_t beg, hts_pos_t end)
 {
     int i, min_off;
     if (p->n == 0) return 0;
-    min_off = bed_minoff(p, beg, end);
+    min_off = bed_minoff(p, beg);
 
     for (i = min_off; i < p->n; ++i) {
-        if ((int)(p->a[i]>>32) >= end) break; // out of range; no need to proceed
-        if ((int32_t)p->a[i] > beg && (int32_t)(p->a[i]>>32) < end)
+        if (p->a[i].beg >= end) break; // out of range; no need to proceed
+        if (p->a[i].end > beg && p->a[i].beg < end)
             return 1; // find the overlap; return
     }
     return 0;
 }
 
-int bed_overlap(const void *_h, const char *chr, int beg, int end)
+int bed_overlap(const void *_h, const char *chr, hts_pos_t beg, hts_pos_t end)
 {
     const reghash_t *h = (const reghash_t*)_h;
     khint_t k;
@@ -202,11 +210,11 @@ void bed_unify(void *reg_hash) {
             continue;
 
         for (new_n = 0, j = 1; j < p->n; j++) {
-            if ((uint32_t)p->a[new_n] < (uint32_t)(p->a[j]>>32)) {
+            if (p->a[new_n].end < p->a[j].beg) {
                 p->a[++new_n] = p->a[j];
             } else {
-                if ((uint32_t)p->a[new_n] < (uint32_t)p->a[j])
-                    p->a[new_n] = (p->a[new_n] & 0xFFFFFFFF00000000) | (uint32_t)(p->a[j]);
+                if (p->a[new_n].end < p->a[j].end)
+                    p->a[new_n].end = p->a[j].end;
             }
         }
 
@@ -257,16 +265,20 @@ void *bed_read(const char *fn)
     if (NULL == h) return NULL;
     // read the list
     fp = strcmp(fn, "-")? gzopen(fn, "r") : gzdopen(fileno(stdin), "r");
-    if (fp == 0) return 0;
+    if (fp == 0) goto fail;
     ks = ks_init(fp);
     if (NULL == ks) goto fail;  // In case ks_init ever gets error checking...
-    while (ks_getuntil(ks, KS_SEP_LINE, &str, &dret) > 0) { // read a line
+    int ks_len;
+    while ((ks_len = ks_getuntil(ks, KS_SEP_LINE, &str, &dret)) >= 0) { // read a line
         char *ref = str.s, *ref_end;
-        unsigned int beg = 0, end = 0;
+        uint64_t beg = 0, end = 0;
         int num = 0;
         khint_t k;
         bed_reglist_t *p;
 
+        if (ks_len == 0)
+            continue; // skip blank lines
+
         line++;
         while (*ref && isspace(*ref)) ref++;
         if ('\0' == *ref) continue;  // Skip blank lines
@@ -275,7 +287,7 @@ void *bed_read(const char *fn)
         while (*ref_end && !isspace(*ref_end)) ref_end++;
         if ('\0' != *ref_end) {
             *ref_end = '\0';  // terminate ref and look for start, end
-            num = sscanf(ref_end + 1, "%u %u", &beg, &end);
+            num = sscanf(ref_end + 1, "%"SCNu64" %"SCNu64, &beg, &end);
         }
         if (1 == num) {  // VCF-style format
             end = beg--; // Counts from 1 instead of 0 for BED files
@@ -293,7 +305,8 @@ void *bed_read(const char *fn)
             } else {
                 fprintf(stderr,
                         "[bed_read] Parse error reading \"%s\" at line %u : "
-                        "end (%u) must not be less than start (%u)\n",
+                        "end (%"PRIu64") must not be less "
+                        "than start (%"PRIu64")\n",
                         fn, line, end, beg);
             }
             errno = 0; // Prevent caller from printing misleading error messages
@@ -318,18 +331,24 @@ void *bed_read(const char *fn)
         // Add begin,end to the list
         if (p->n == p->m) {
             p->m = p->m ? p->m<<1 : 4;
-            p->a = realloc(p->a, p->m * sizeof(uint64_t));
-            if (NULL == p->a) goto fail;
+            hts_pair_pos_t *new_a = realloc(p->a, p->m * sizeof(p->a[0]));
+            if (NULL == new_a) goto fail;
+            p->a = new_a;
         }
-        p->a[p->n++] = (uint64_t)beg<<32 | end;
+        p->a[p->n].beg = beg;
+        p->a[p->n++].end = end;
     }
     // FIXME: Need to check for errors in ks_getuntil.  At the moment it
     // doesn't look like it can return one.  Possibly use gzgets instead?
 
+    if (gzclose(fp) != Z_OK) {
+        fp = NULL;
+        goto fail;
+    }
+    if (bed_index(h) != 0)
+        goto fail;
     ks_destroy(ks);
-    gzclose(fp);
     free(str.s);
-    bed_index(h);
     //bed_unify(h);
     return h;
  fail:
@@ -361,7 +380,7 @@ void bed_destroy(void *_h)
     kh_destroy(reg, h);
 }
 
-static void *bed_insert(void *reg_hash, char *reg, unsigned int beg, unsigned int end) {
+static void *bed_insert(void *reg_hash, char *reg, hts_pos_t beg, hts_pos_t end) {
 
     reghash_t *h;
     khint_t k;
@@ -390,10 +409,12 @@ static void *bed_insert(void *reg_hash, char *reg, unsigned int beg, unsigned in
     // Add beg and end to the list
     if (p->n == p->m) {
         p->m = p->m ? p->m<<1 : 4;
-        p->a = realloc(p->a, p->m * sizeof(uint64_t));
-        if (NULL == p->a) goto fail;
+        hts_pair_pos_t *new_a = realloc(p->a, p->m * sizeof(p->a[0]));
+        if (NULL == new_a) goto fail;
+        p->a = new_a;
     }
-    p->a[p->n++] = (uint64_t)beg<<32 | end;
+    p->a[p->n].beg = beg;
+    p->a[p->n++].end = end;
 
 fail:
     return h;
@@ -413,10 +434,10 @@ static void *bed_filter(void *reg_hash, void *tmp_hash) {
     reghash_t *t;
     bed_reglist_t *p, *q;
     khint_t l, k;
-    uint64_t *new_a;
+    hts_pair_pos_t *new_a;
     int i, j, new_n, min_off;
     const char *reg;
-    uint32_t beg, end;
+    hts_pos_t beg, end;
 
     h = (reghash_t *)reg_hash;
     t = (reghash_t *)tmp_hash;
@@ -434,20 +455,21 @@ static void *bed_filter(void *reg_hash, void *tmp_hash) {
         if (k == kh_end(h) || !(p = &kh_val(h, k)) || !(p->n))
             continue;
 
-        new_a = (uint64_t *)calloc(q->n + p->n, sizeof(uint64_t));
+        new_a = calloc(q->n + p->n, sizeof(new_a[0]));
         if (!new_a)
             return NULL;
         new_n = 0;
 
         for (i = 0; i < q->n; i++) {
-            beg = (uint32_t)(q->a[i]>>32);
-            end = (uint32_t)(q->a[i]);
+            beg = q->a[i].beg;
+            end = q->a[i].end;
 
-            min_off = bed_minoff(p, beg, end);
+            min_off = bed_minoff(p, beg);
             for (j = min_off; j < p->n; ++j) {
-                if ((uint32_t)(p->a[j]>>32) >= end) break; // out of range; no need to proceed
-                if ((uint32_t)(p->a[j]) > beg && (uint32_t)(p->a[j]>>32) < end) {
-                    new_a[new_n++] = ((uint64_t)MAX((uint32_t)(p->a[j]>>32), beg) << 32) | MIN((uint32_t)p->a[j], end);
+                if (p->a[j].beg >= end) break; // out of range; no need to proceed
+                if (p->a[j].end > beg && p->a[j].beg < end) {
+                    new_a[new_n].beg = MAX(p->a[j].beg, beg);
+                    new_a[new_n++].end = MIN(p->a[j].end, end);
                 }
             }
         }
@@ -494,6 +516,11 @@ void *bed_hash_regions(void *reg_hash, char **regs, int first, int last, int *op
 
     for (i=first; i<last; i++) {
 
+        // Note, ideally we would call sam_parse_region here, but it's complicated by not
+        // having the sam header known and the likelihood of the bed file containing data for other
+        // references too which we currently just ignore.
+        //
+        // TO DO...
         q = hts_parse_reg(regs[i], &beg, &end);
         if (q) {
             if ((int)(q - regs[i] + 1) > 1024) {
@@ -514,28 +541,43 @@ void *bed_hash_regions(void *reg_hash, char **regs, int first, int last, int *op
 
         //if op==1 insert reg to the bed hash table
         if (*op && !(bed_insert(h, reg, beg, end))) {
-            fprintf(stderr, "Error when inserting region='%s' in the bed hash table at address=%p!\n", regs[i], h);
+            fprintf(stderr, "Error when inserting region='%s' in the bed hash table at address=%p!\n", regs[i], (void *)h);
         }
         //if op==0, first insert the regions in the temporary hash table,
         //then filter the bed hash table using it
         if (!(*op) && !(bed_insert(t, reg, beg, end))) {
-            fprintf(stderr, "Error when inserting region='%s' in the temporary hash table at address=%p!\n", regs[i], t);
+            fprintf(stderr, "Error when inserting region='%s' in the temporary hash table at address=%p!\n", regs[i], (void *)t);
         }
     }
 
     if (!(*op)) {
-        bed_index(t);
+        if (bed_index(t) != 0)
+            goto fail;
         bed_unify(t);
-        h = bed_filter(h, t);
+        if (bed_filter(h, t) == NULL)
+            goto fail;
         bed_destroy(t);
+        t = NULL;
     }
 
     if (h) {
-        bed_index(h);
+        if (bed_index(h) != 0)
+            goto fail;
         bed_unify(h);
     }
 
     return h;
+
+ fail:
+    // Clean up whichever hash we made
+    if (reg_hash) {
+        if (t)
+            bed_destroy(t);
+    } else {
+        if (h)
+            bed_destroy(h);
+    }
+    return NULL;
 }
 
 const char* bed_get(void *reg_hash, int i, int filter) {
@@ -553,6 +595,14 @@ const char* bed_get(void *reg_hash, int i, int filter) {
     return kh_key(h, i);
 }
 
+/**
+ * Create a region list from a the region hash table
+ * @param  reg_hash  The region hash table
+ * @param  filter    0 - allow all regions, 1 - allow only selected regions
+ * @param  n_reg     Pointer to the returned region number
+ * @return           The regions list as a hts_reglist_t
+ */
+
 hts_reglist_t *bed_reglist(void *reg_hash, int filter, int *n_reg) {
 
     reghash_t *h;
@@ -596,8 +646,8 @@ hts_reglist_t *bed_reglist(void *reg_hash, int filter, int *n_reg) {
         reglist[count].max_end = 0;
 
         for (j = 0; j < p->n; j++) {
-            reglist[count].intervals[j].beg = (uint32_t)(p->a[j]>>32);
-            reglist[count].intervals[j].end = (uint32_t)(p->a[j]);
+            reglist[count].intervals[j].beg = p->a[j].beg;
+            reglist[count].intervals[j].end = p->a[j].end;
 
             if (reglist[count].intervals[j].end > reglist[count].max_end)
                 reglist[count].max_end = reglist[count].intervals[j].end;
