@@ -6,6 +6,7 @@ import collections
 import struct
 import copy
 import array
+from pysam import CDEL, CDIFF, CEQUAL, CINS, CMATCH, CPAD, CREF_SKIP, CSOFT_CLIP
 
 from TestUtils import (
     checkFieldEqual,
@@ -49,6 +50,38 @@ class TestAlignedSegment(ReadTest):
     and manipulated.
     """
 
+    def check_get_aligned_pairs_combos(self, a, exp):
+        def positions(exp):  return [(ppos, rpos)        for ppos, rpos, base, cigar in exp]
+        def with_seq(exp):   return [(ppos, rpos, base)  for ppos, rpos, base, cigar in exp]
+        def with_cigar(exp): return [(ppos, rpos, cigar) for ppos, rpos, base, cigar in exp]
+
+        self.assertEqual(a.get_aligned_pairs(), positions(exp))
+        self.assertEqual(a.get_aligned_pairs(with_seq=True), with_seq(exp))
+        self.assertEqual(a.get_aligned_pairs(with_cigar=True), with_cigar(exp))
+        self.assertEqual(a.get_aligned_pairs(with_seq=True, with_cigar=True), exp)
+
+        exp = [(ppos, rpos, base, cigar) for ppos, rpos, base, cigar in exp if ppos is not None and rpos is not None]
+
+        self.assertEqual(a.get_aligned_pairs(matches_only=True), positions(exp))
+        self.assertEqual(a.get_aligned_pairs(matches_only=True, with_seq=True), with_seq(exp))
+        self.assertEqual(a.get_aligned_pairs(matches_only=True, with_cigar=True), with_cigar(exp))
+        self.assertEqual(a.get_aligned_pairs(matches_only=True, with_seq=True, with_cigar=True), exp)
+
+    def check_get_aligned_pairs_combos_without_MD(self, a, exp):
+        def positions(exp): return [(ppos, rpos) for ppos, rpos, cigar in exp]
+
+        self.assertEqual(a.get_aligned_pairs(), positions(exp))
+        with self.assertRaises(ValueError): a.get_aligned_pairs(with_seq=True)
+        self.assertEqual(a.get_aligned_pairs(with_cigar=True), exp)
+        with self.assertRaises(ValueError): a.get_aligned_pairs(with_seq=True, with_cigar=True)
+
+        exp = [(ppos, rpos, cigar) for ppos, rpos, cigar in exp if ppos is not None and rpos is not None]
+
+        self.assertEqual(a.get_aligned_pairs(matches_only=True), positions(exp))
+        with self.assertRaises(ValueError): a.get_aligned_pairs(matches_only=True, with_seq=True)
+        self.assertEqual(a.get_aligned_pairs(matches_only=True, with_cigar=True), exp)
+        with self.assertRaises(ValueError): a.get_aligned_pairs(matches_only=True, with_seq=True, with_cigar=True)
+
     def testEmpty(self):
 
         a = pysam.AlignedSegment()
@@ -313,50 +346,50 @@ class TestAlignedSegment(ReadTest):
             ],
         )
 
-        self.assertEqual(
-            a.get_aligned_pairs(),
+        self.check_get_aligned_pairs_combos_without_MD(
+            a,
             [
-                (0, 20),
-                (1, 21),
-                (2, 22),
-                (3, 23),
-                (4, 24),
-                (5, 25),
-                (6, 26),
-                (7, 27),
-                (8, 28),
-                (9, 29),
-                (None, 30),
-                (10, 31),
-                (11, 32),
-                (12, 33),
-                (13, 34),
-                (14, 35),
-                (15, 36),
-                (16, 37),
-                (17, 38),
-                (18, 39),
-                (19, None),
-                (20, 40),
-                (21, 41),
-                (22, 42),
-                (23, 43),
-                (24, 44),
-                (25, 45),
-                (26, 46),
-                (27, 47),
-                (28, 48),
-                (29, 49),
-                (30, 50),
-                (31, 51),
-                (32, 52),
-                (33, 53),
-                (34, 54),
-                (35, 55),
-                (36, 56),
-                (37, 57),
-                (38, 58),
-                (39, 59),
+                (0, 20, CMATCH),
+                (1, 21, CMATCH),
+                (2, 22, CMATCH),
+                (3, 23, CMATCH),
+                (4, 24, CMATCH),
+                (5, 25, CMATCH),
+                (6, 26, CMATCH),
+                (7, 27, CMATCH),
+                (8, 28, CMATCH),
+                (9, 29, CMATCH),
+                (None, 30, CDEL),
+                (10, 31, CMATCH),
+                (11, 32, CMATCH),
+                (12, 33, CMATCH),
+                (13, 34, CMATCH),
+                (14, 35, CMATCH),
+                (15, 36, CMATCH),
+                (16, 37, CMATCH),
+                (17, 38, CMATCH),
+                (18, 39, CMATCH),
+                (19, None, CINS),
+                (20, 40, CMATCH),
+                (21, 41, CMATCH),
+                (22, 42, CMATCH),
+                (23, 43, CMATCH),
+                (24, 44, CMATCH),
+                (25, 45, CMATCH),
+                (26, 46, CMATCH),
+                (27, 47, CMATCH),
+                (28, 48, CMATCH),
+                (29, 49, CMATCH),
+                (30, 50, CMATCH),
+                (31, 51, CMATCH),
+                (32, 52, CMATCH),
+                (33, 53, CMATCH),
+                (34, 54, CMATCH),
+                (35, 55, CMATCH),
+                (36, 56, CMATCH),
+                (37, 57, CMATCH),
+                (38, 58, CMATCH),
+                (39, 59, CMATCH),
             ],
         )
 
@@ -435,40 +468,24 @@ class TestAlignedSegment(ReadTest):
     def test_get_aligned_pairs_soft_clipping(self):
         a = self.build_read()
         a.cigartuples = ((4, 2), (0, 35), (4, 3))
-        self.assertEqual(
-            a.get_aligned_pairs(),
-            [(0, None), (1, None)]
+        self.check_get_aligned_pairs_combos_without_MD(
+            a,
+            [(0, None, CSOFT_CLIP), (1, None, CSOFT_CLIP)]
             + [
-                (qpos, refpos)
+                (qpos, refpos, CMATCH)
                 for (qpos, refpos) in zip(range(2, 2 + 35), range(20, 20 + 35))
             ]
-            + [(37, None), (38, None), (39, None)],
-        )
-        self.assertEqual(
-            a.get_aligned_pairs(True),
-            # [(0, None), (1, None)] +
-            [
-                (qpos, refpos)
-                for (qpos, refpos) in zip(range(2, 2 + 35), range(20, 20 + 35))
-            ]
-            # [(37, None), (38, None), (39, None)]
+            + [(37, None, CSOFT_CLIP), (38, None, CSOFT_CLIP), (39, None, CSOFT_CLIP)],
         )
 
     def test_get_aligned_pairs_hard_clipping(self):
         a = self.build_read()
         a.cigartuples = ((5, 2), (0, 35), (5, 3))
-        self.assertEqual(
-            a.get_aligned_pairs(),
+        self.check_get_aligned_pairs_combos_without_MD(
+            a,
             # No seq, no seq pos
             [
-                (qpos, refpos)
-                for (qpos, refpos) in zip(range(0, 0 + 35), range(20, 20 + 35))
-            ],
-        )
-        self.assertEqual(
-            a.get_aligned_pairs(True),
-            [
-                (qpos, refpos)
+                (qpos, refpos, CMATCH)
                 for (qpos, refpos) in zip(range(0, 0 + 35), range(20, 20 + 35))
             ],
         )
@@ -476,23 +493,12 @@ class TestAlignedSegment(ReadTest):
     def test_get_aligned_pairs_skip(self):
         a = self.build_read()
         a.cigarstring = "2M100D38M"
-        self.assertEqual(
-            a.get_aligned_pairs(),
-            [(0, 20), (1, 21)]
-            + [(None, refpos) for refpos in range(22, 22 + 100)]
+        self.check_get_aligned_pairs_combos_without_MD(
+            a,
+            [(0, 20, CMATCH), (1, 21, CMATCH)]
+            + [(None, refpos, CDEL) for refpos in range(22, 22 + 100)]
             + [
-                (qpos, refpos)
-                for (qpos, refpos) in zip(
-                    range(2, 2 + 38), range(20 + 2 + 100, 20 + 2 + 100 + 38)
-                )
-            ],
-        )
-        self.assertEqual(
-            a.get_aligned_pairs(True),
-            [(0, 20), (1, 21)] +
-            # [(None, refpos) for refpos in range(21, 21+100)] +
-            [
-                (qpos, refpos)
+                (qpos, refpos, CMATCH)
                 for (qpos, refpos) in zip(
                     range(2, 2 + 38), range(20 + 2 + 100, 20 + 2 + 100 + 38)
                 )
@@ -502,17 +508,10 @@ class TestAlignedSegment(ReadTest):
     def test_get_aligned_pairs_match_mismatch(self):
         a = self.build_read()
         a.cigartuples = ((7, 20), (8, 20))
-        self.assertEqual(
-            a.get_aligned_pairs(),
+        self.check_get_aligned_pairs_combos_without_MD(
+            a,
             [
-                (qpos, refpos)
-                for (qpos, refpos) in zip(range(0, 0 + 40), range(20, 20 + 40))
-            ],
-        )
-        self.assertEqual(
-            a.get_aligned_pairs(True),
-            [
-                (qpos, refpos)
+                (qpos, refpos, CEQUAL if qpos < 20 else CDIFF)
                 for (qpos, refpos) in zip(range(0, 0 + 40), range(20, 20 + 40))
             ],
         )
@@ -522,8 +521,16 @@ class TestAlignedSegment(ReadTest):
         a.cigartuples = ((0, 1), (6, 1), (0, 1))
         # The padding operation is like an insertion into the reference.
         # See comment in test_get_aligned_pairs_padding_with_seq (below).
-        self.assertEqual(a.get_aligned_pairs(),
-                         [(0, 20), (1, None), (2, 21)])
+        self.check_get_aligned_pairs_combos_without_MD(a,
+                         [(0, 20, CMATCH), (1, None, CPAD), (2, 21, CMATCH)])
+
+    def test_get_aligned_pairs_padding_via_cigarstring(self):
+        a = self.build_read()
+        a.cigarstring = "1M1P1M"
+        # The padding operation is like an insertion into the reference.
+        # See comment in test_get_aligned_pairs_padding_with_seq (below).
+        self.check_get_aligned_pairs_combos_without_MD(a,
+                         [(0, 20, CMATCH), (1, None, CPAD), (2, 21, CMATCH)])
 
     def test_get_aligned_pairs_padding_with_seq(self):
         a = self.build_read()
@@ -550,80 +557,80 @@ class TestAlignedSegment(ReadTest):
         # string: "Alternatively, to describe the same alignments, we can
         # modify the reference sequence to contain pads that make room for
         # sequences inserted relative to the reference."
-        self.assertEqual(a.get_aligned_pairs(with_seq=True),
-                         [(0, 20, 'A'), (1, None, None), (2, 21, 'T')])
+        self.check_get_aligned_pairs_combos(a,
+            [(0, 20, "A", CMATCH), (1, None, None, CPAD), (2, 21, "T", CMATCH)])
 
     def test_get_aligned_pairs(self):
         a = self.build_read()
         a.query_sequence = "A" * 9
         a.cigarstring = "9M"
         a.set_tag("MD", "9")
-        self.assertEqual(
-            a.get_aligned_pairs(with_seq=True),
+        self.check_get_aligned_pairs_combos(
+            a,
             [
-                (0, 20, "A"),
-                (1, 21, "A"),
-                (2, 22, "A"),
-                (3, 23, "A"),
-                (4, 24, "A"),
-                (5, 25, "A"),
-                (6, 26, "A"),
-                (7, 27, "A"),
-                (8, 28, "A"),
+                (0, 20, "A", CMATCH),
+                (1, 21, "A", CMATCH),
+                (2, 22, "A", CMATCH),
+                (3, 23, "A", CMATCH),
+                (4, 24, "A", CMATCH),
+                (5, 25, "A", CMATCH),
+                (6, 26, "A", CMATCH),
+                (7, 27, "A", CMATCH),
+                (8, 28, "A", CMATCH),
             ],
         )
 
         a.set_tag("MD", "4C4")
-        self.assertEqual(
-            a.get_aligned_pairs(with_seq=True),
+        self.check_get_aligned_pairs_combos(
+            a,
             [
-                (0, 20, "A"),
-                (1, 21, "A"),
-                (2, 22, "A"),
-                (3, 23, "A"),
-                (4, 24, "c"),
-                (5, 25, "A"),
-                (6, 26, "A"),
-                (7, 27, "A"),
-                (8, 28, "A"),
+                (0, 20, "A", CMATCH),
+                (1, 21, "A", CMATCH),
+                (2, 22, "A", CMATCH),
+                (3, 23, "A", CMATCH),
+                (4, 24, "c", CMATCH),
+                (5, 25, "A", CMATCH),
+                (6, 26, "A", CMATCH),
+                (7, 27, "A", CMATCH),
+                (8, 28, "A", CMATCH),
             ],
         )
 
         a.cigarstring = "5M2D4M"
         a.set_tag("MD", "4C^TT4")
-        self.assertEqual(
-            a.get_aligned_pairs(with_seq=True),
+        self.check_get_aligned_pairs_combos(
+            a,
             [
-                (0, 20, "A"),
-                (1, 21, "A"),
-                (2, 22, "A"),
-                (3, 23, "A"),
-                (4, 24, "c"),
-                (None, 25, "T"),
-                (None, 26, "T"),
-                (5, 27, "A"),
-                (6, 28, "A"),
-                (7, 29, "A"),
-                (8, 30, "A"),
+                (0, 20, "A", CMATCH),
+                (1, 21, "A", CMATCH),
+                (2, 22, "A", CMATCH),
+                (3, 23, "A", CMATCH),
+                (4, 24, "c", CMATCH),
+                (None, 25, "T", CDEL),
+                (None, 26, "T", CDEL),
+                (5, 27, "A", CMATCH),
+                (6, 28, "A", CMATCH),
+                (7, 29, "A", CMATCH),
+                (8, 30, "A", CMATCH),
             ],
         )
 
         a.cigarstring = "5M2D2I2M"
         a.set_tag("MD", "4C^TT2")
-        self.assertEqual(
-            a.get_aligned_pairs(with_seq=True),
+        self.check_get_aligned_pairs_combos(
+            a,
             [
-                (0, 20, "A"),
-                (1, 21, "A"),
-                (2, 22, "A"),
-                (3, 23, "A"),
-                (4, 24, "c"),
-                (None, 25, "T"),
-                (None, 26, "T"),
-                (5, None, None),
-                (6, None, None),
-                (7, 27, "A"),
-                (8, 28, "A"),
+                (0, 20, "A", CMATCH),
+                (1, 21, "A", CMATCH),
+                (2, 22, "A", CMATCH),
+                (3, 23, "A", CMATCH),
+                (4, 24, "c", CMATCH),
+                (None, 25, "T", CDEL),
+                (None, 26, "T", CDEL),
+                (5, None, None, CINS),
+                (6, None, None, CINS),
+                (7, 27, "A", CMATCH),
+                (8, 28, "A", CMATCH),
             ],
         )
 
@@ -643,54 +650,21 @@ class TestAlignedSegment(ReadTest):
         a.cigarstring = "5M1N5M"
         a.set_tag("MD", "10")
 
-        self.assertEqual(
-            a.get_aligned_pairs(with_seq=True),
+        self.check_get_aligned_pairs_combos(
+            a,
             [
-                (0, 20, "A"),
-                (1, 21, "A"),
-                (2, 22, "A"),
-                (3, 23, "A"),
-                (4, 24, "A"),
-                (None, 25, None),
-                (5, 26, "A"),
-                (6, 27, "A"),
-                (7, 28, "A"),
-                (8, 29, "A"),
-                (9, 30, "A"),
-            ],
-        )
-
-        self.assertEqual(
-            a.get_aligned_pairs(with_seq=False),
-            [
-                (0, 20),
-                (1, 21),
-                (2, 22),
-                (3, 23),
-                (4, 24),
-                (None, 25),
-                (5, 26),
-                (6, 27),
-                (7, 28),
-                (8, 29),
-                (9, 30),
-            ],
-        )
-
-        self.assertEqual(
-            a.get_aligned_pairs(matches_only=True, with_seq=False),
-            [
-                (0, 20),
-                (1, 21),
-                (2, 22),
-                (3, 23),
-                (4, 24),
-                (5, 26),
-                (6, 27),
-                (7, 28),
-                (8, 29),
-                (9, 30),
-            ],
+                (0, 20, "A", CMATCH),
+                (1, 21, "A", CMATCH),
+                (2, 22, "A", CMATCH),
+                (3, 23, "A", CMATCH),
+                (4, 24, "A", CMATCH),
+                (None, 25, None, CREF_SKIP),
+                (5, 26, "A", CMATCH),
+                (6, 27, "A", CMATCH),
+                (7, 28, "A", CMATCH),
+                (8, 29, "A", CMATCH),
+                (9, 30, "A", CMATCH),
+            ]
         )
 
     def test_equivalence_matches_only_and_with_seq(self):
@@ -698,32 +672,22 @@ class TestAlignedSegment(ReadTest):
         a.query_sequence = "ACGT" * 2
         a.cigarstring = "4M1D4M"
         a.set_tag("MD", "4^x4")
-        full = (
-            list(zip(range(0, 4), range(20, 24), "ACGT"))
-            + [(None, 24, "x")]
-            + list(zip(range(4, 8), range(25, 29), "ACGT"))
-        )
-        self.assertEqual(a.get_aligned_pairs(matches_only=False, with_seq=True), full)
-
-        self.assertEqual(
-            a.get_aligned_pairs(matches_only=True, with_seq=True),
-            [x for x in full if x[0] is not None and x[1] is not None],
+        self.check_get_aligned_pairs_combos(
+            a,
+            list(zip(range(0, 4), range(20, 24), "ACGT", [CMATCH] * 4))
+            + [(None, 24, "x", CDEL)]
+            + list(zip(range(4, 8), range(25, 29), "ACGT", [CMATCH] * 4)),
         )
 
         a = self.build_read()
         a.query_sequence = "ACGT" * 2
         a.cigarstring = "4M1N4M"
         a.set_tag("MD", "8")
-        full = (
-            list(zip(range(0, 4), range(20, 24), "ACGT"))
-            + [(None, 24, None)]
-            + list(zip(range(4, 8), range(25, 29), "ACGT"))
-        )
-        self.assertEqual(a.get_aligned_pairs(matches_only=False, with_seq=True), full)
-
-        self.assertEqual(
-            a.get_aligned_pairs(matches_only=True, with_seq=True),
-            [x for x in full if x[0] is not None and x[1] is not None],
+        self.check_get_aligned_pairs_combos(
+            a,
+            list(zip(range(0, 4), range(20, 24), "ACGT", [CMATCH] * 4))
+            + [(None, 24, None, 3)]
+            + list(zip(range(4, 8), range(25, 29), "ACGT", [CMATCH] * 4)),
         )
 
     def test_get_aligned_pairs_lowercase_md(self):
@@ -1069,7 +1033,7 @@ class TestBaseModifications(unittest.TestCase):
         expect = {
             ("C", 0, "m"): [(6, 102), (17, 128), (20, 153), (31, 179), (34, 204)],
             ("N", 0, "n"): [(15, 212)],
-            ("C", 0, 76792): [(19, 161), (34, 187)],
+            ("C", 0, 76792): [(19, 161), (34, 33)],
         }
 
         with pysam.AlignmentFile(filename, check_sq=False) as inf:
@@ -1090,6 +1054,21 @@ class TestBaseModifications(unittest.TestCase):
             r = next(iter(inf))
             self.assertDictEqual(r.modified_bases, expect)
 
+    def testExplicit(self):
+        """reference bases should always be the same nucleotide
+        """
+        filename = os.path.join(BAM_DATADIR, "MM-explicit.bam")
+        expected_output = [
+            {("C", 0, "m"): [(9, 200), (10, 50), (14, 160)], ("C", 0, "h"): [(9, 10), (10, 170), (14, 20)]},
+            {("C", 0, "m"): [(9, 200), (10, 50), (13, 10), (14, 160), (16, 10)],
+             ("C", 0, "h"): [(9, 10), (10, 170), (13, 5), (14, 20), (16, 5)]},
+            {("C", 0, "m"): [(9, 200), (14, 160)], ("C", 0, "h"): [(9, 10), (10, 170), (13, 5), (14, 20), (16, 5)]},
+        ]
+
+        with pysam.AlignmentFile(filename, check_sq=False) as inf:
+            for r, expected in zip(inf, expected_output):
+                self.assertDictEqual(r.modified_bases, expected)
+
     def testMulti(self):
         """reference bases should always be the same nucleotide
         """
@@ -1609,6 +1588,18 @@ class TestSetTagGetTag(ReadTest):
             alt_value_type="I",
         )
 
+    def test_set_tag_invalid_value_type(self):
+        with self.assertRaises(ValueError):
+            self.check_tag("TT", "abc", value_type="#")
+
+    def test_set_array_tag_invalid_value_type(self):
+        with self.assertRaises(ValueError):
+            self.check_tag("TT", array.array('I', range(4)), value_type='#')
+
+    def test_set_array_tag_invalid_typecode(self):
+        with self.assertRaises(ValueError):
+            self.check_tag("TT", array.array('L', range(4)), value_type=None)
+
 
 class TestSetTagsGetTag(TestSetTagGetTag):
     def check_tag(self, tag, value, value_type, alt_value_type=None):
